unit VideoPlayer;

interface

uses
  Windows, ActiveX, MMSystem, SysUtils, DShow,
  RiffWalk;

const
  IID_IMediaDet: TGUID = '{65BD0710-24D2-4FF7-9324-ED2E5D3ABAFA}';
  CLSID_MediaDet: TGUID = '{65BD0711-24D2-4FF7-9324-ED2E5D3ABAFA}';

type
  // DShow.pasɂȂC^[tF[X
  ISampleGrabberCB = interface(IUnknown)
    ['{0579154A-2B53-4994-B0D0-E773148EFF85}']
    function SampleCB(SampleTime: Double; pSample: IMediaSample): HResult; stdcall;
    function BufferCB(SampleTime: Double; pBuffer: PByte; BufferLen: longint): HResult; stdcall;
  end;
  ISampleGrabber = interface(IUnknown)
    ['{6B652FFF-11FE-4FCE-92AD-0266B5D7C78F}']
    function SetOneShot(OneShot: BOOL): HResult; stdcall;
    function SetMediaType(var pType: TAM_MEDIA_TYPE): HResult; stdcall;
    function GetConnectedMediaType(out pType: TAM_MEDIA_TYPE): HResult; stdcall;
    function SetBufferSamples(BufferThem: BOOL): HResult; stdcall;
    function GetCurrentBuffer(var pBufferSize: longint; pBuffer: Pointer): HResult; stdcall;
    function GetCurrentSample(out ppSample: IMediaSample): HResult; stdcall;
    function SetCallback(pCallback: ISampleGrabberCB; WhichMethodToCallback: longint): HResult; stdcall;
  end;
  IMediaDet = interface(IUnknown)
    ['{65BD0710-24D2-4FF7-9324-ED2E5D3ABAFA}']
    function get_Filter(out pVal: IUnknown): HResult; stdcall;
    function put_Filter(newVal: IUnknown): HResult; stdcall;
    function get_OutputStreams(out pVal: longint): HResult; stdcall;
    function get_CurrentStream(out pVal: longint): HResult; stdcall;
    function put_CurrentStream(newVal: longint): HResult; stdcall;
    function get_StreamType(out pVal: TGUID): HResult; stdcall;
    function get_StreamTypeB(out pVal: WideString): HResult; stdcall;
    function get_StreamLength(out pVal: Double): HResult; stdcall;
    function get_Filename(out pVal: WideString): HResult; stdcall;
    function put_Filename(pVal: WideString): HResult; stdcall;
    function GetBitmapBits(streamTime: Double; var pBufferSize: longint; var pBuffer: Pchar;
      Width: longint; Height: longint): HResult; stdcall;
    function WriteBitmapBits(streamTime: Double; Width: longint; Height: longint;
      Filename: WideString): HResult; stdcall;
    function get_StreamMediaType(out pVal: TAM_MEDIA_TYPE): HResult; stdcall;
    function GetSampleGrabber(out ppVal: ISampleGrabber): HResult; stdcall;
    function get_FrameRate(out pVal: Double): HResult; stdcall;
    function EnterBitmapGrabMode(SeekTime: Double): HResult; stdcall;
  end;

  TVPModes = (vpPlay, vpPause, vpStop);

  TVideoPlayer = class(TObject)
  protected
    Figb: IGraphBuilder;
    Fimc: IMediaControl;
    Fims: IMediaSeeking;
    Fime: IMediaEvent;
    Fibv: IBasicVideo;
    //Fivf: IVideoFrameStep;
    Fiba: IBasicAudio;
    Fivw: IVideoWindow;
    TimerID: Integer;
    Duration: Comp;
    function GetVolume: Integer;
    procedure SetVolume(Value: Integer);
    function GetPosition: Integer;
    procedure SetPosition(Value: Integer);
  public
    Mode: TVpModes;
    FileName: string;
    Width: Integer;
    Height: Integer;
    Length: Integer;   // msec
    Loaded: Boolean;
    DisplayWnd: HWND;
    Fps: Double;
    AVIMainHeader: TAVIMainHeader;
    BitmapInfoHeader: TBitmapInfoHeader;
    WaveFormat: TWaveFormat;
    OnTimer: procedure of object;
    OnEndPlay: procedure of object;
    property Volume: Integer read GetVolume write SetVolume;
    property Position: Integer read GetPosition write SetPosition; // msec
    procedure FreeInterface;
    procedure LoadFile(const FileName: string);
    procedure Play;
    procedure Stop;
    procedure Pause;
    procedure PauseOnly;
    procedure SetWindowPosition(ALeft, ATop, AWidth, AHeight: Longint);
    function WriteBitmapBits(const FileName: string): Boolean;
  end;

implementation

procedure TimeProc(IDEvent, uReserved, dwUser, dwReserved1, dwReserved2: Integer); stdcall;
begin
  with TVideoPlayer(dwUser) do
  begin
    if Loaded and Assigned(OnTimer) then
      OnTimer;
  end;
end;

function TVideoPlayer.GetVolume: Integer;
begin
  if Loaded then
    Fiba.get_Volume(Result)
  else
    Result := 0;
end;

procedure TVideoPlayer.SetVolume(Value: Integer);
begin
  if Loaded then Fiba.put_Volume(Value);
end;

function TVideoPlayer.GetPosition: Integer;
var
  cp: Comp;
begin
  if Loaded then
  begin
    Fims.GetCurrentPosition(cp);
    if cp >= Duration then
    begin
      Fimc.Stop; // ̂ȁ[
      Mode := vpStop;
      if Assigned(OnEndPlay) then
        OnEndPlay;
    end;
    Result := Trunc(cp / 10000);
  end else
    Result := 0;
end;

procedure TVideoPlayer.SetPosition(Value: Integer);
var
  CurPos, StopPos: Comp;
begin
  if Loaded then
  begin
    CurPos := Value * 10000.0;
    if CurPos >= Duration then CurPos := Duration;
    StopPos := 0;
    Fims.SetPositions(CurPos, AM_SEEKING_AbsolutePositioning,
      StopPos, AM_SEEKING_NoPositioning);
    GetPosition;
  end;
end;

procedure TVideoPlayer.FreeInterface;
begin
  if Assigned(Fimc) then
    Fimc.Stop;
  if Assigned(Fivw) then
  begin
    Fivw.put_Visible(False);
    Fivw.put_Owner(OAHWND(0));
  end;
  Fimc := nil;
  Fivw := nil;
  Fims := nil;
  Figb := nil;
  Fibv := nil;
  //Fivf := nil;
  Fiba := nil;
  Fime := nil;
  Loaded := False;
end;

procedure TVideoPlayer.Play;
var
  cp: Comp;
begin
  if Loaded then
  begin
    Fims.GetCurrentPosition(cp);
    if (cp = Duration) and (Mode = vpStop) then
      Position := 0;
    Fimc.Run;
    Mode := vpPlay;
    if TimerID <> 0 then
    begin
      timeKillEvent(TimerID);
      TimerID := 0;
    end;
    TimerID := timeSetEvent(100, 50, TimeProc, DWORD(Self), TIME_PERIODIC);
  end;
end;

procedure TVideoPlayer.Stop;
begin
  if Loaded then
  begin
    Fimc.Stop;
    Position := 0;
    Fimc.StopWhenReady;
    Mode := vpStop;
  end;
end;

procedure TVideoPlayer.PauseOnly;
begin
  if Loaded then
  begin
    timeKillEvent(TimerID);
    TimerID := 0;
    Fimc.Pause;
    Mode := vpPause;
  end;
end;

procedure TVideoPlayer.Pause;
begin
  if Mode = vpPause then
    Play
  else
    PauseOnly;
end;

procedure TVideoPlayer.LoadFile(const FileName: string);
var
  tpf: Double;
  F: File;
  ws: WideString;
  size: Integer;
begin
  Loaded := False;
  FillChar(AVIMainHeader, SizeOf(AVIMainHeader), 0);
  FillChar(BitmapInfoHeader, SizeOf(BitmapInfoHeader), 0);
  FillChar(WaveFormat, SizeOf(WaveFormat), 0);
  if Assigned(Fimc) then
  begin
    Fimc.Stop;
    FreeInterface;
  end;
  FileMode := 0;
  AssignFile(F, FileName);
  try
    Reset(F, 1);
  except
    Exit;
  end;

  try
    if SearchChunk(F, 'RIFF', size) then
    begin
      SearchChunk(F, 'avih', size);
      BlockRead(F, AVIMainHeader, SizeOf(AVIMainHeader));
      SearchChunk(F, 'strf', size);
      Seek(F, FilePos(F)+8);
      BlockRead(F, BitmapInfoHeader, SizeOf(BitmapInfoHeader));
      SearchChunk(F, 'strf', size);
      Seek(F, FilePos(F)+8);
      BlockRead(F, WaveFormat, SizeOf(WaveFormat));
    end;
  except
  end;
  CloseFile(F);

  try
    CoCreateInstance(CLSID_FilterGraph, nil, CLSCTX_INPROC_SERVER,
      IID_IGraphBuilder, Figb);
    ws := FileName;
    if Figb.RenderFile(PWideChar(ws), nil) <> S_OK then
    begin
      FreeInterface;
      Exit;
    end;

    Fivw := Figb as IVideoWindow;
    Fivw.put_Owner(OAHWND(DisplayWnd));
    Fivw.put_WindowStyle(WS_CHILD or WS_CLIPSIBLINGS);
    Fivw.put_MessageDrain(DisplayWnd);
    Fimc := Figb as IMediaControl;
    Fims := Figb as IMediaSeeking;
    Fime := Figb as IMediaEvent;
    //Fims.SetTimeFormat(TIME_FORMAT_FRAME);
    Fims.SetTimeFormat(TIME_FORMAT_MEDIA_TIME);
    Fibv := Figb as IBasicVideo;
    Fibv.GetVideoSize(Width, Height);
    Fibv.get_AvgTimePerFrame(tpf);
    if tpf = 0 then
    begin
      if AVIMainHeader.dwMicroSecPerFrame > 0 then
        Fps := 1000000 / AVIMainHeader.dwMicroSecPerFrame
      else
        Fps := 0;
    end else
      Fps := 1 / tpf;
    Fims.GetDuration(Duration);
    Length := Trunc(Duration / 10000);
    Self.FileName := FileName;
    Position := 0;
    Loaded := True;
    //Fivf := Figb as IVideoFrameStep;
    //Fims.SetRate(2.0);
    Fiba := Figb as IBasicAudio;
  except
    FreeInterface;
  end;
end;

procedure TVideoPlayer.SetWindowPosition(ALeft, ATop, AWidth, AHeight: Longint);
begin
  Fivw.SetWindowPosition(ALeft, ATop, AWidth, AHeight);
end;

function TVideoPlayer.WriteBitmapBits(const FileName: string): Boolean;
var
  ws: WideString;
  i, n: Integer;
  imd: IMediaDet;
  mt: TGUID;
begin
  Result := False;
  if Loaded then
  begin
    if FAILED(CoCreateInstance(CLSID_MediaDet, nil, CLSCTX_INPROC_SERVER, IID_IMediaDet, imd)) then
      Exit;
    ws := Self.FileName;
    if FAILED(imd.put_Filename(ws)) then
      Exit;
    if FAILED(imd.get_OutputStreams(n)) then
      Exit;
    for i := 0 to n - 1 do
    begin
      if FAILED(imd.put_CurrentStream(i)) then
        Continue;
      if FAILED(imd.get_StreamType(mt)) then
        Continue;
      if CompareMem(@mt, @MEDIATYPE_Video, SizeOf(mt)) then
      begin
        Result := imd.WriteBitmapBits(Position / 1000.0, Width, Height, FileName)= S_OK;
        Break;
      end;
    end;
  end;
end;

end.
